#ifndef _UNION_FIND_H
#define _UNION_FIND_H

//STL
#include <ext/hash_map>
//TPIE
#include <ami_stream.h>

template<class TIME, class ID>
class uf_t {
public:
	TIME time;
	ID id;
	friend ostream& operator<< (ostream &ostr, const uf_t &e) {
		return ostr << e.time << " " << e.id;
	}
};

template<class ID>
class rep_t {
public:
	ID id;
	ID rep;
};


template<class ID>
class Node {  // in-memory union-find node
public:
	ID id;
	Node<ID> *parent;
  int height;
	ID rep;

	Node(ID i)
	{
		id = rep = i;
		parent = this;
		height = 0;
	}
}; 

template<class ID>
class OnlineUF {
private:
	bool largest;

public:
  __gnu_cxx::hash_map<ID, Node<ID>*> *id2node;
	Node<ID> * find(Node<ID> *p);

	OnlineUF(bool iflargest = false) 
	{
		id2node = new __gnu_cxx::hash_map<ID, Node<ID>*>;
		largest = iflargest;
	}

	~OnlineUF() 
	{
		typename __gnu_cxx::hash_map<ID, Node<ID>*>::iterator it0;
		it0 = id2node->begin();
		while (it0 != id2node->end()) {
			delete it0->second;
			it0++;
		}
		delete id2node;
	}

	ID Find(ID x);

	void Union(vector<ID>& ids);
};

/*
	-- Input --
	
	in: a stream of (time, id) pairs, sorted by time.  A number of
	consecutive pairs with the same time makes up one operation.  If
	there is only pair, it is a find operation, with id being the
	operand.  If there are more than one pairs, it is a union operation
	that joins the id's in the pairs.  in will be deleted upon
	returning.  The stream file on disk might also get changed.
	So in general in should be PERSIST_DELETE, and should not be used 
	after the function call.

	It is assumed that operators < and == are implemented for class TIME
	and ID.  The function object hash<T> must have a specialization for
	T=ID.

	out: a stream of id's, each answering a find query in the input
	stream in the order given.
*/

template<class TIME, class ID>
void union_find(AMI_STREAM<uf_t<TIME,ID> > *in, AMI_STREAM<ID> *out, 
								bool largest = false);

// if largest = true, return the largest as the representative

template<class TIME, class ID>
void uf_mm(AMI_STREAM<uf_t<TIME,ID> > *in, AMI_STREAM<ID> *out, 
					 AMI_STREAM<rep_t<ID> > *replist, bool largest = false);

#include "union_find_impl.h"

#endif
